;;; gim-mouse.el --- mouse support within the text terminal

;;; Copyright (C) 1994,1995 Alessandro Rubini <rubini@linux.it>
;;;               parts are by Ian T Zimmermann <itz@rahul.net>, 1995,1998
;;;               more changes for gim, Ian Zimmerman <itz@speakeasy.org> 2002

;; Maintainer: Ian Zimmerman <itz@speakeasy.org>
;; Keywords: mouse gim linux

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;; This library provides access to mouse events as reported by the
;; gim package. It uses the program "gimev" to get mouse events.
;; It tries to reproduce the functionality offered by emacs under X.
;; The "gim" server runs under Linux, so this package is rather
;; Linux-dependent.

;; Developed for GNU Emacs 20.7, likely won't work with many others
;; too much internals dependent cruft here.

;; To use, put something like this in your .emacs:
;;
;; (and (not window-system)
;;      (menu-bar-mode -1)  ;gim-mouse is incompatible with menu-bar, use context menus
;;      (load-library "gim-mouse")
;;      (gim-mouse-run))

(defconst gim-mouse-rcsid "$Id: gim-mouse.el,v 1.8 2002/09/12 02:36:07 itz Exp $")


(require 'advice)

(defgroup gim-mouse nil
  "Use your mouse in Emacs on the Linux virtual terminals."
  :group 'mouse
  :prefix "gim-mouse-")

(defcustom gim-mouse-device "mouse"
  "The name of the device which provides mouse events.
Please read the `gim' distribution documentation for more info."
  :group 'gim-mouse
  :type 'string)

(defcustom gim-mouse-ps-command "ps h %d"
  "Shell command to execute to read emacs' process table entry.
It should contain exactly one %d conversion, which will be matched to emacs pid." 
  :group 'gim-mouse
  :type 'string)

(defcustom gim-mouse-ps-output-regexp " *[0-9]+ +tty\\([0-9]+\\)"
  "Regexp matching the output of `gim-mouse-ps-command'.
Its first group should match emacs' controlling tty number.  For example,
if emacs is running on /dev/tty2, the first group should match \"2\"."
  :group 'gim-mouse
  :type 'regexp)

(defcustom gim-mouse-gimev-extra-args nil
  "List of extra arguments (strings) to pass to the `gimev' subprocess.
For example, you can use this to change the default set of accepted
keyboard modifiers, which includes all modifier keys.
Please see `gim' documentation for more info."
  :group 'gim-mouse
  :type '(repeat string))

(defvar gim-mouse-process nil 
  "Embeds the process which passes mouse events to emacs.
It is used by the program gim-mouse.")

(defvar gim-mouse-filter-accumulator ""
  "Accumulates input from the mouse reporting process.")

(defvar gim-mouse-current-xy '(0 . 0)
  "Stores the last mouse position gim-mouse has been told about.")

(defvar gim-mouse-drag-start nil
  "Stores the coordinates where a drag event started.
Whenever a drag starts in a special part of a window (not the text),
the `translated' starting coordinates including the window and part
involved are saved here.  This is necessary lest they get
re-translated when the button goes up, at which time window
configuration may have changed.")

(defvar gim-mouse-prev-set-selection-function 'x-set-selection)
(defvar gim-mouse-prev-get-selection-function 'x-get-selection)


;;; Code:

;; get the number of the current virtual console

(defun gim-mouse-tty ()
  "Returns number of virtual terminal Emacs is running on, as a string.
For example, \"2\" for /dev/tty2."
  (let ((buffer (generate-new-buffer " *gim-mouse*"))
        (proc-reader (format gim-mouse-ps-command (emacs-pid))))
    (prog1 (save-excursion
             (set-buffer buffer)
             (goto-char (point-min))
             (shell-command proc-reader t)
             (and (looking-at gim-mouse-ps-output-regexp)
                  (match-string 1)))
      (kill-buffer buffer))))


;; due to a horrible kludge in Emacs' keymap handler
;; (read_key_sequence) mouse clicks on funny parts of windows generate
;; TWO events, the first being a dummy of the sort '(mode-line).
;; That's why Per Abrahamsen's code in xt-mouse.el doesn't work for
;; the modeline, for instance.

;; now get this:  the Emacs C code that generates these fake events
;; depends on certain things done by the very lowest level input
;; handlers; namely the symbols for the events (for instance
;; 'C-S-double-mouse-2) must have an 'event-kind property, set to
;; 'mouse-click.  Since events from unread-command-events do not pass
;; through the low level handlers, they don't get this property unless
;; I set it myself.  I imagine this has caused innumerable attempts by
;; hackers to do things similar to gim-mouse to lose.

;; The next page of code is devoted to fixing this ugly problem.

;; WOW! a fully general powerset generator
;; (C) Ian Zimmerman Mon Mar 23 12:00:16 PST 1998 :-)
(defun gim-mouse-powerset (l)
  (if (null l) '(nil)
    (let ((l1 (gim-mouse-powerset (cdr l)))
          (first (nth 0 l)))
      (append
       (mapcar (function (lambda (l) (cons first l))) l1) l1))))

;; and a slightly less general cartesian product
(defun gim-mouse-cartesian (l1 l2)
  (if (null l1) l2
    (append (mapcar (function (lambda (x) (append (nth 0 l1) x))) l2)
            (gim-mouse-cartesian (cdr l1) l2))))
      
(let* ((modifier-sets (gim-mouse-powerset '(control meta shift)))
       (typed-sets (gim-mouse-cartesian '((down) (drag))
                                      '((mouse-1) (mouse-2) (mouse-3))))
       (multipled-sets (gim-mouse-cartesian '((double) (triple)) typed-sets))
       (all-sets (gim-mouse-cartesian modifier-sets multipled-sets)))
  (while all-sets
    (let ((event-sym (event-convert-list (nth 0 all-sets))))
      (if (not (get event-sym 'event-kind))
          (put event-sym 'event-kind 'mouse-click)))
    (setq all-sets (cdr all-sets))))


;;; This fun is partly Copyright (C) 1994 Per Abrahamsen <abraham@iesd.auc.dk>
;; This is basically a feeble attempt to mimic what the c function
;; buffer_posn_from_coords in dispnew.c does.  I wish that function
;; were exported to Lisp.

(defun gim-mouse-lispy-buffer-posn-from-coords (w col line)
  "Return buffer position of character at COL and LINE within window W.
COL and LINE are glyph coordinates, relative to W topleft corner."
  (save-window-excursion
    (select-window w)
    (save-excursion
      (move-to-window-line line)
      (move-to-column (+ col (current-column)
                         (if (not (window-minibuffer-p w)) 0
                           (- (minibuffer-prompt-width)))
                         (max 0 (1- (window-hscroll)))))
      (point))))

;; compute one element of the form (WINDOW BUFFERPOS (COL . ROW) TIMESTAMP)

(defun gim-mouse-make-event-element (x-dot-y-avec-time)
  (let* ((x-dot-y (nth 0 x-dot-y-avec-time))
         (x (car x-dot-y))
         (y (cdr x-dot-y))
         (timestamp (nth 1 x-dot-y-avec-time))
         (w (window-at x y))
         (left-top-right-bottom (window-edges w))
         (left (nth 0 left-top-right-bottom))
         (top (nth 1 left-top-right-bottom))
         (right (nth 2 left-top-right-bottom))
         (bottom (nth 3 left-top-right-bottom))
         (coords-or-part (coordinates-in-window-p x-dot-y w)))
    (cond
     ((consp coords-or-part)
      (let ((wx (car coords-or-part)) (wy (cdr coords-or-part)))
        (if (< wx (- right left 1))
            (list w 
                  (gim-mouse-lispy-buffer-posn-from-coords w wx wy)
                  coords-or-part timestamp)
          (list w 'vertical-scroll-bar
                (cons (1+ wy) (- bottom top)) timestamp))))
     ((eq coords-or-part 'mode-line)
      (list w 'mode-line (cons (- x left) 0) timestamp))
     ((eq coords-or-part 'vertical-line)
      (list w 'vertical-line (cons 0 (- y top)) timestamp)))))     

;; completely rewritten for gim by itz Sat Jun  8 23:00:31 PDT 2002, using matching,
;; because it seems the Emacs reader changed enough to break the convenient Lisp form way

(defun gim-mouse-make-event ()
  "Makes a Lisp style event from the contents of mouse input accumulator.
Also trims the accumulator by all the data used to build the event."
  (save-match-data
    (and
     (string-match
      "^\\([-0-9a-z]+\\) +\\(-?[0-9]+\\) +\\(-?[0-9]+\\) +\\(-?[0-9]+\\)\\( +\\([1-3]\\)\\( +\\(-?[0-9]+\\) +\\(-?[0-9]+\\) +\\(-?[0-9]+\\)\\)?\\)? *\n"
      gim-mouse-filter-accumulator)
     (let* ((event-type (intern (match-string 1 gim-mouse-filter-accumulator)))
            (current-x (string-to-int (match-string 2 gim-mouse-filter-accumulator)))
            (current-y (string-to-int (match-string 3 gim-mouse-filter-accumulator)))
            (current-xy (cons current-x current-y))
            (current-time (string-to-int (match-string 4 gim-mouse-filter-accumulator)))
            (current-xy-avec-time (list current-xy current-time)))
       (prog1
           (cond
            ;;sink all events on the stupid text mode menubar.
            ((and menu-bar-mode (eq 0 (cdr gim-mouse-current-xy))) nil)
        
            ;;drag
            ((match-string 7 gim-mouse-filter-accumulator) 
             (let* ((count (string-to-int (match-string 6 gim-mouse-filter-accumulator)))
                    (start-x (string-to-int (match-string 8 gim-mouse-filter-accumulator)))
                    (start-y (string-to-int (match-string 9 gim-mouse-filter-accumulator)))
                    (start-time (string-to-int (match-string 10 gim-mouse-filter-accumulator)))
                    (start-xy-avec-time (list (cons start-x start-y) start-time))
                    (start-element
                     (or gim-mouse-drag-start
                         (gim-mouse-make-event-element start-xy-avec-time)))
                    (end-element
                     (gim-mouse-make-event-element current-xy-avec-time)))
               (setq gim-mouse-drag-start nil)
               (list event-type start-element end-element count)))

            ;;down or up
            ((match-string 5 gim-mouse-filter-accumulator)
             (let ((count (string-to-int (match-string 6 gim-mouse-filter-accumulator)))
                   (element
                    (gim-mouse-make-event-element current-xy-avec-time)))
               (if (and (not gim-mouse-drag-start)
                        (symbolp (nth 1 element)))
                   ;; OUCH! GOTCHA! emacs uses setc[ad]r on these!
                   (setq gim-mouse-drag-start (copy-sequence element))
                 (setq gim-mouse-drag-start nil))
               (list event-type element count)))

            ;;movement
            (t
             (list (if (eq 'vertical-scroll-bar
                        (nth 1 gim-mouse-drag-start)) 'scroll-bar-movement
                     'mouse-movement)
                   (gim-mouse-make-event-element current-xy-avec-time))))

         ;;store current position for mouse-position
         (setq gim-mouse-current-xy current-xy)
         
         ;;trim accumulator by the prefix we just processed
         (setq gim-mouse-filter-accumulator
               (substring gim-mouse-filter-accumulator (match-end 0))))))))

(defun gim-mouse-process-filter (proc string)
  (setq gim-mouse-filter-accumulator
        (concat gim-mouse-filter-accumulator string))
  (let ((event (gim-mouse-make-event)))
    (while event
      (if (or track-mouse 
              (not (eq 'mouse-movement (event-basic-type event))))
          (setq unread-command-events
                (nconc unread-command-events (list event))))
      (setq event (gim-mouse-make-event)))))


;; this overrides a C function which stupidly assumes (no X => no mouse)
(defadvice mouse-position (around gim-mouse-mouse-position activate)
  "Return the gim-mouse-position unless running with a window system.
The (secret) scrollbar interface is not implemented yet."
  (if (not window-system)
      (setq ad-return-value
            (cons (selected-frame) gim-mouse-current-xy))
    ad-do-it))

(setq mouse-sel-set-selection-function
      (function (lambda (type value)
                  (if (not window-system)
                      (if (eq 'PRIMARY type) (kill-new value))
                    (funcall gim-mouse-prev-set-selection-function
                             type value)))))

(setq mouse-sel-get-selection-function
      (function (lambda (type)
                  (if (not window-system)
                      (if (eq 'PRIMARY type)
                          (current-kill 0) "")
                    (funcall gim-mouse-prev-get-selection-function type)))))

(add-hook 'suspend-hook
          (function (lambda ()
                      (and gim-mouse-process
                           (process-send-string
                            gim-mouse-process "stop\n")))))

(add-hook 'suspend-resume-hook
          (function (lambda ()
                      (and gim-mouse-process
                           (process-send-string gim-mouse-process "start\n")))))


;;; User commands

(defun gim-mouse-quit ()
  "Stop getting mouse events from an asynchronous process."
  (interactive)
  (process-send-string gim-mouse-process "quit\n")
  (delete-process gim-mouse-process)
  (setq gim-mouse-process nil))

(defun gim-mouse-run ()
  "Starts getting a stream of mouse events from an asynchronous process.
Only works if Emacs is running on a virtual terminal without a window system.
Returns the newly created asynchronous process."
  (interactive)
  (let ((tty (gim-mouse-tty))
        (process-connection-type t))
    (if (or window-system (not (stringp tty)))
        (error "Run gim-mouse on a virtual terminal without a window system"))
    (setq gim-mouse-process
          (apply 'start-process "gim-mouse" nil
                 "gimev" "-d" gim-mouse-device "-t" tty "-P"
                 gim-mouse-gimev-extra-args)))
  (setq gim-mouse-filter-accumulator "")
  (set-process-filter gim-mouse-process 'gim-mouse-process-filter)
  (process-kill-without-query gim-mouse-process)
  gim-mouse-process)

(provide 'gim-mouse)

;;; gim-mouse.el ends here
